#pragma once

#include <cstdint>
#include <Pokitto.h>
#include <LibLog>
#include "images/grass.h"
#include "images/sand.h"
#include "images/line.h"
#include "images/asphalt.h"

template<typename SDF_T>
class SDFRenderer {
public:
    static SDFRenderer& get_instance() {
        static SDFRenderer self;
        return self;
    }
    
    void set_distance_bitmap(const SDF_T* bitmap, std::uint32_t sw, std::uint32_t sh, std::uint32_t dw, std::uint32_t dh) {
        _bitmap = bitmap;
        _bitmap_w = sw;
        _bitmap_h = sh;
        
        _w = dw;
        _h = dh;
        
        _step_u = (static_cast<std::int64_t>(sw - 1) << Q) / dw;
        _step_v = (static_cast<std::int64_t>(sh - 1) << Q) / dh;
    }
    
    void shift_distance_field(std::int32_t x, std::int32_t y) {
        _x = x;
        _y = y;
    }
    
    static void line_filler(std::uint8_t* line, std::uint32_t y, bool skip) {
        if(y == 0) get_instance().frame_start();
        
        if(!skip) get_instance().draw_line(line, y);
    }

private:
    static constexpr std::uint32_t Q = 16;
    static constexpr std::int32_t ONE = 1 << Q;
    
    static constexpr int SCREEN_W = PROJ_LCDWIDTH;
    static constexpr int SCREEN_H = PROJ_LCDHEIGHT;
    
    constexpr SDFRenderer() {}
    
    // Interpolation
    static constexpr std::int32_t mix(std::int32_t a, std::int32_t b, std::int32_t f) { return a + ((f * (b - a)) >> Q); }
    
    static constexpr std::int32_t distance_field_shader(std::int32_t dist, std::int32_t x, std::int32_t y) {
        // Divide the distance into 16 bands
        std::int32_t i = (8 + (dist >> (Q - 3))) & 15;
        
        // Map input parameters to pixel color using textures or solid colors
        // return _tex[i][2 + (x & 63) + 64 * (y & 63)]; // 26 fps
        return _pal[i]; // 32 fps
    }
    
    void frame_start() {
        _start_x = (_x > 0) ? (_x < SCREEN_W ? 0 : SCREEN_W - _x) : -_x;
        _end_x = (_x + _w) < SCREEN_W ? _w : SCREEN_W - _x;
        
        _start_y = (_y > 0) ? (_y < SCREEN_H ? _y : SCREEN_H) : 0;
        _end_y = (_y + _h) < SCREEN_H ? (_y + _h) : SCREEN_H;
    }
    
    void draw_line(std::uint8_t* line, std::int32_t y) {
        // Check if the current line is within the bounds of the distance field
        if((y >= _start_y) && (y < _end_y)) {
            // Adjust x-coordinate relative to the starting point of the distance field and 
            // shift line pointer accordingly
            std::int32_t x = -_x;
            line += _x;
            
            // Fill the empty space before the left edge of the distance field with the background color
            while(x < _start_x) {
                line[x] = _bg_color;
                ++x;
            }
            
            // Adjust y-coordinate relative to the starting point of the distance field
            y -= _y;
            
            std::int32_t v = _step_v * y + (_step_v >> 1);
            std::int32_t v0 = v >> Q;
            
            // Fetch the two rows from the SDF bitmap that contribute to the current line
            const std::int8_t* top_row = &_bitmap[v0 * _bitmap_w];
            const std::int8_t* bot_row = &_bitmap[(v0 + 1) * _bitmap_w];
            
            std::int32_t u = _step_u * x + (_step_u >> 1);
            std::int32_t u0 = u >> Q;
            
            // Read the pixels from SDF bitmap that contribute to the top_left and bottom_left corners
            std::int32_t top = top_row[u0];
            std::int32_t bot = bot_row[u0];
            
            // Calculate the interpolated value between top_left and bottom_left pixels
            std::int32_t vf = v & (ONE - 1);
            std::int32_t mix_r = mix(top, bot, vf);
            
            std::int32_t u_end = _step_u * _end_x + (_step_u >> 1);
            
            // Draw a line of the actual distance field
            while(u < u_end) {
                // The old interpolated right side value becomes the new left side value
                std::int32_t mix_l = mix_r;
                
                u0 = u >> Q;
                
                // Read the pixels from SDF bitmap that contribute to the top_right and bottom_right corners
                top = top_row[u0 + 1];
                bot = bot_row[u0 + 1];
                
                // Calculate the interpolated value between top_right and bottom_right pixels
                mix_r = mix(top, bot, vf);
                
                // u_break is the horizontal position where we enter to the next SDF bitmap pixels and need to read new top and bottom values
                std::int32_t u_break = (u0 + 1) << Q;
                if(u_break > u_end) u_break = u_end;
                
                // Loop as long as we can use the same interpolated mix_l and mix_r values
                while(u < u_break) {
                    std::int32_t dist = mix(mix_l, mix_r, u & (ONE - 1));
                    
                    // Normalize to the range -1...1 as Q16 fixed point number
                    dist *= (1 << Q) / 127;
                    
                    // Get pixel color from distance value and x, y coordinates
                    line[x] = distance_field_shader(dist, x, y);
                    ++x;
                    
                    u += _step_u;
                }
            }
            
            // Fill the empty space after the right edge of the distance field with the background color
            while(x < SCREEN_W - _x) {
                line[x] = _bg_color;
                ++x;
            }
        }
        else {
            // Fill the entire line with the background color if current line is outside the bounds of the distance field
            for(std::int32_t x = 0; x < SCREEN_W; ++x) {
                line[x] = _bg_color;
            }
        }
    }
    
    std::int32_t _bitmap_w = 0;
    std::int32_t _bitmap_h = 0;
    const SDF_T* _bitmap = nullptr;
    
    std::uint8_t _bg_color = 36;
    static inline const std::uint8_t _pal[16] = {115, 115, 115, 115, 115, 100, 100, 6, 3, 3, 3, 3, 3, 3, 3, 3};
    static inline const std::uint8_t* _tex[16] = {
        grass, grass, grass, grass, grass, grass, sand, line, 
        asphalt, asphalt, asphalt, asphalt, asphalt, asphalt, asphalt, asphalt
    };
    
    std::int16_t _x = 0;
    std::int16_t _y = 0;
    std::int16_t _w = 0;
    std::int16_t _h = 0;
    
    std::int32_t _step_u = 0;
    std::int32_t _step_v = 0;
    
    std::int32_t _start_x = 0;
    std::int32_t _end_x = 0;
    std::int32_t _start_y = 0;
    std::int32_t _end_y = 0;
};
