#!/usr/bin/env python3
# Convert a black and white image to a distance field image

import argparse
from PIL import Image
from scipy.ndimage import distance_transform_edt

# Create the parser and add arguments
parser = argparse.ArgumentParser(description='Convert a black and white image to a distance field image.')
parser.add_argument('in_image', metavar='input_image', type=str, help='the path to the input image file')
parser.add_argument('out_image', metavar='output_image', type=str, help='the path to the output image file')
parser.add_argument('-d', '--distance', type=int, required=True, help='specifies the range of the calculated distance field in pixels')
parser.add_argument('-u', '--unsigned', action='store_true', help='calculate distance values only for the outside of the shape')

# Parse the arguments
args = parser.parse_args()

# Open and convert the image to grayscale
image = Image.open(args.in_image).convert('L')

# Convert the grayscale image to a binary image
bin_image = [[image.getpixel((x, y)) > 128 for x in range(image.size[0])] for y in range(image.size[1])]

# Generate the SDF for the shape and its inverse
sdf_positive = distance_transform_edt(bin_image)
sdf_negative = distance_transform_edt([[not pixel for pixel in row] for row in bin_image])

# Subtract the negative SDF from the positive SDF to create a signed SDF
sdf = [[max(0.5, pos) - max(0.5, neg) for pos, neg in zip(pos_row, neg_row)] for pos_row, neg_row in zip(sdf_positive, sdf_negative)]

# Normalize the SDF
sdf_normalized = [[max(-1.0, min(1.0, val / args.distance)) for val in row] for row in sdf]

# Save the SDF as an image
sdf_image = Image.new('L', image.size)
if args.unsigned:
  sdf_image.putdata([int(val * 255) for row in sdf_normalized for val in row])
else:
  sdf_image.putdata([int(128 + val * 127) for row in sdf_normalized for val in row])
sdf_image.save(args.out_image)

